import 'reflect-metadata';

/**
 * Decorator that helps you implement incomplete multiple inheritance.
 *
 * Mixin classes do not support CIs because they are executed earlier. Used to refine the behavior
 * of other classes, not intended to spawn self-used objects. Mixins are useful for bringing common
 * functionality into a single place and being reusable.
 *
 * @usageNotes
 *
 * The parent class needs to implement the mixin classes (only the declaration of properties or methods)
 * so that the project can be built without errors at the compilation stage.
 *
 * If the mixin class has private fields or methods that you do not want
 * to declare in the parent class, then you can create your own type/interface and declare it:
 *
 * If you want to pass parameters to the constructor of the mixin class,
 * you need to pass an array where the first element will be the mixin class itself,
 * and all subsequent elements are its parameters.
 *
 * @example
 * ```ts
 * class SubscriptionMixin {
 *    // ...
 * }
 *
 * class GoodbyeMixin {
 *    // ...
 * }
 *
 * class GreeterMixin {
 *    public get fullName(): string {
 *      return `${this.firstName} ${this.surName}`;
 *    }
 *
 *    public constructor(private firstName: string, private surName: string) {
 *    }
 *
 *    public greet(): string {
 *      return `Hello, ${this.firstName} ${this.surName}!`;
 *    }
 * }
 *
 * type GreeterMixinType = Omit<GreeterMixin, 'firstName' | 'surName'>;
 * // or
 * interface GreeterInterface extends Omit<GreeterMixin, 'firstName' | 'surName'> {}
 *
 * @Component({...})
 * @Mixins([
 *    SubscriptionMixin,
 *    GoodbyeMixin,
 *    [GreeterMixin, 'John', 'Wick'],
 * ])
 * class HelloWorld implements GreeterMixinType {
 *    public fullName: string;
 *    public greet: () => string;
 *    // Other declarations: SubscriptionMixin, GoodbyeMixin
 *
 *    public constructor() {
 *      alert(this.greet());
 *    }
 * }
 * ```
 *
 * @Annotation
 */
const Mixins = (bases) => {
    return (target) => {
        for (const base of bases) {
            try {
                let constructor;
                let instance;
                if (Array.isArray(base)) {
                    const [mixinCtor, ...params] = base;
                    constructor = mixinCtor;
                    instance = new constructor(...params);
                }
                else {
                    constructor = base;
                    instance = new base();
                }
                mappingPropertyFromInstance(target, instance);
                mappingProperty(target, constructor);
            }
            catch (e) {
                console.warn(e);
            }
        }
    };
};
function mappingPropertyFromInstance(target, instance) {
    Object.defineProperties(target.prototype, Object.getOwnPropertyDescriptors(instance));
}
function mappingProperty(target, baseCtor) {
    Object
        .getOwnPropertyNames(baseCtor.prototype)
        .filter((name) => name !== 'constructor')
        .forEach((name) => {
        const descriptor = Object.getOwnPropertyDescriptor(baseCtor.prototype, name);
        try {
            // Deep copy object fields.
            descriptor.value = JSON.parse(JSON.stringify(descriptor.value));
        }
        catch (e) {
        }
        Object.defineProperty(target.prototype, name, descriptor);
    });
}

const DEFAULT_OPTS = {
    size: 20,
    duration: 0,
};
/**
 * Higher-order function that memoizes functions.
 *
 * @see `{@link MemorizeOptions}`
 */
const memorizeFn = (fn, opts = DEFAULT_OPTS) => {
    const cache = new Map();
    if (opts.duration) {
        setInterval(() => cache.clear(), opts.duration);
    }
    return (...args) => {
        const key = JSON.stringify(args);
        if (!cache.has(key)) {
            cache.set(key, fn(...args));
        }
        try {
            return cache.get(key);
        }
        finally {
            if (cache.size > opts.size) {
                cache.delete(cache.keys().next().value);
            }
        }
    };
};

/**
 * A decorator that memoizes class methods.
 *
 * @see `{@link MemorizeOptions}`
 *
 * @Annotation
 */
function Memorize(opts) {
    return (target, name, descriptor) => {
        descriptor.value = memorizeFn(target[name], opts);
        return descriptor;
    };
}

/**
 * The decorator overrides the properties in the calling context
 * and executes the method, then returns the previous property values.
 */
const OverrideProps = (props) => {
    return (target, key, descriptor) => {
        return Object.assign(Object.assign({}, descriptor), { value: function wrapper() {
                let prevProps = {};
                Object
                    .keys(props)
                    .forEach((k) => {
                    prevProps[k] = this[k];
                    this[k] = props[k];
                });
                const method = descriptor.value.apply(this, arguments);
                Object
                    .keys(prevProps)
                    .forEach((k) => {
                    this[k] = prevProps[k];
                    prevProps = null;
                });
                return method;
            } });
    };
};

var RuntimeTypeLevel;
(function (RuntimeTypeLevel) {
    RuntimeTypeLevel["Log"] = "Log";
    RuntimeTypeLevel["Warn"] = "Warn";
    RuntimeTypeLevel["Error"] = "Error";
    RuntimeTypeLevel["Throw"] = "Throw";
    RuntimeTypeLevel["Hidden"] = "Hidden";
})(RuntimeTypeLevel || (RuntimeTypeLevel = {}));

const composeMessage = ({ prop, expect, actual }) => {
    return `The parameter '${prop}' must be of type '${expect}', but not the type '${actual}'.`;
};
const levelStrategies = {
    [RuntimeTypeLevel.Log]: (error) => console.log(composeMessage(error)),
    [RuntimeTypeLevel.Warn]: (error) => console.warn(composeMessage(error)),
    [RuntimeTypeLevel.Error]: (error) => console.error(composeMessage(error)),
    [RuntimeTypeLevel.Hidden]: (error) => { },
    [RuntimeTypeLevel.Throw]: (error) => {
        throw new Error(composeMessage(error));
    },
};

const RUNTIME_CHECK_OPTIONS_KEY = Symbol('__pm_runtime_check_options_key__');
const RUNTIME_CHECK_PARAM_KEY = Symbol('__pm_runtime_check_type_key__');
const RUNTIME_CHECK_OPTIONS_TARGET = class TargetOptions {
};
const ARGUMENT_NAMES = /([^\s,]+)/g;
const STRIP_COMMENTS = /(\/\/.*$)|(\/\*[\s\S]*?\*\/)|(\s*=[^,\)]*(('(?:\\'|[^'\r\n])*')|("(?:\\"|[^"\r\n])*"))|(\s*=[^,\)]*))/mg;
const getTypeName = (type) => {
    if (typeof type === 'function' && type.name) {
        return type.name;
    }
    else if (null !== type && typeof type === 'object' && 'constructor' in type) {
        return type.constructor.name;
    }
    return Object.prototype.toString.call(type).slice(8, -1);
};
const getParamNames = (func) => {
    const fnStr = func.toString().replace(STRIP_COMMENTS, '');
    return fnStr.slice(fnStr.indexOf('(') + 1, fnStr.indexOf(')')).match(ARGUMENT_NAMES) || [];
};
const setGlobalRuntimeTypeOptions = (options = {}) => {
    const opts = Reflect.getOwnMetadata(RUNTIME_CHECK_OPTIONS_KEY, RUNTIME_CHECK_OPTIONS_TARGET) || {};
    Reflect.defineMetadata(RUNTIME_CHECK_OPTIONS_KEY, Object.assign(Object.assign({}, opts), options), RUNTIME_CHECK_OPTIONS_TARGET);
};
const validate = (prop, propValue, type) => {
    const actual = getTypeName(propValue);
    const expect = getTypeName(type);
    const isInstance = propValue instanceof type;
    if (!isInstance && actual !== expect) {
        return { prop, expect, actual };
    }
    return null;
};
const outputErrorMessage = (opts, error) => {
    const levelOutputStrategy = levelStrategies[opts.level];
    if (levelOutputStrategy) {
        levelOutputStrategy(error);
    }
    else {
        console.warn(`Level: No strategy found '${opts.level}'!`);
    }
};

const T = (type) => {
    return (target, propName, index) => {
        if (typeof index !== 'undefined') {
            const params = Reflect.getOwnMetadata(RUNTIME_CHECK_PARAM_KEY, target, propName) || [];
            params.push({ index, type });
            Reflect.defineMetadata(RUNTIME_CHECK_PARAM_KEY, params, target, propName);
        }
    };
};

const RuntimeType = (options = {}) => {
    const globalOptions = Reflect.getOwnMetadata(RUNTIME_CHECK_OPTIONS_KEY, RUNTIME_CHECK_OPTIONS_TARGET) || {};
    const opts = Object.assign(Object.assign({ level: RuntimeTypeLevel.Error }, globalOptions), options);
    return (target, methodName, descriptor) => {
        const paramNames = getParamNames(target[methodName]);
        const params = Reflect.getOwnMetadata(RUNTIME_CHECK_PARAM_KEY, target, methodName) || [];
        return Object.assign(Object.assign({}, descriptor), { value: (...args) => {
                for (const { index, type } of params.reverse()) {
                    const paramValue = args[index];
                    const paramName = paramNames[index];
                    const error = validate(paramName, paramValue, type);
                    if (error) {
                        outputErrorMessage(opts, error);
                    }
                }
                descriptor.value.apply(target, args);
            } });
    };
};

/*
 * Public API Surface of decorators
 */

/**
 * Generated bundle index. Do not edit.
 */

export { ARGUMENT_NAMES, Memorize, Mixins, OverrideProps, RUNTIME_CHECK_OPTIONS_KEY, RUNTIME_CHECK_OPTIONS_TARGET, RUNTIME_CHECK_PARAM_KEY, RuntimeType, RuntimeTypeLevel, STRIP_COMMENTS, T, getParamNames, getTypeName, levelStrategies, memorizeFn, outputErrorMessage, setGlobalRuntimeTypeOptions, validate };
//# sourceMappingURL=plus-minus-decorators.js.map
